@@ -1,10 +1,10 @@
 /* vi: set sw=4 ts=4: */
-/* uncompress for busybox -- (c) 2002 Robert Griebl
+/*
+ * uncompress for busybox -- (c) 2002 Robert Griebl
  *
  * based on the original compress42.c source
  * (see disclaimer below)
  */
-
 /* (N)compress42.c - File compression ala IEEE Computer, Mar 1992.
  *
  * Authors:
@@ -21,9 +21,7 @@
  * marc@suse.de : a small security fix for a buffer overflow
  *
  * [... History snipped ...]
- *
  */
-
 #include "libbb.h"
 #include "bb_archive.h"
 
@@ -73,12 +71,11 @@
  */
 
 IF_DESKTOP(long long) int FAST_FUNC
-unpack_Z_stream(transformer_aux_data_t *aux, int src_fd, int dst_fd)
+unpack_Z_stream(transformer_state_t *xstate)
 {
 	IF_DESKTOP(long long total_written = 0;)
 	IF_DESKTOP(long long) int retval = -1;
 	unsigned char *stackp;
-	long code;
 	int finchar;
 	long oldcode;
 	long incode;
@@ -103,7 +100,7 @@ unpack_Z_stream(transformer_aux_data_t *aux, int src_fd, int dst_fd)
 	/* block compress mode -C compatible with 2.0 */
 	int block_mode; /* = BLOCK_MODE; */
 
-	if (check_signature16(aux, src_fd, COMPRESS_MAGIC))
+	if (check_signature16(xstate, COMPRESS_MAGIC))
 		return -1;
 
 	inbuf = xzalloc(IBUFSIZ + 64);
@@ -115,7 +112,7 @@ unpack_Z_stream(transformer_aux_data_t *aux, int src_fd, int dst_fd)
 
 	/* xread isn't good here, we have to return - caller may want
 	 * to do some cleanup (e.g. delete incomplete unpacked file etc) */
-	if (full_read(src_fd, inbuf, 1) != 1) {
+	if (full_read(xstate->src_fd, inbuf, 1) != 1) {
 		bb_error_msg("short read");
 		goto err;
 	}
@@ -143,8 +140,10 @@ unpack_Z_stream(transformer_aux_data_t *aux, int src_fd, int dst_fd)
 	/* As above, initialize the first 256 entries in the table. */
 	/*clear_tab_prefixof(); - done by xzalloc */
 
-	for (code = 255; code >= 0; --code) {
-		tab_suffixof(code) = (unsigned char) code;
+	{
+		int i;
+		for (i = 255; i >= 0; --i)
+			tab_suffixof(i) = (unsigned char) i;
 	}
 
 	do {
@@ -165,7 +164,7 @@ unpack_Z_stream(transformer_aux_data_t *aux, int src_fd, int dst_fd)
 		}
 
 		if (insize < (int) (IBUFSIZ + 64) - IBUFSIZ) {
-			rsize = safe_read(src_fd, inbuf + insize, IBUFSIZ);
+			rsize = safe_read(xstate->src_fd, inbuf + insize, IBUFSIZ);
 			if (rsize < 0)
 				bb_error_msg_and_die(bb_msg_read_error);
 			insize += rsize;
@@ -175,6 +174,8 @@ unpack_Z_stream(transformer_aux_data_t *aux, int src_fd, int dst_fd)
 				  (insize << 3) - (n_bits - 1));
 
 		while (inbits > posbits) {
+			long code;
+
 			if (free_ent > maxcode) {
 				posbits =
 					((posbits - 1) +
@@ -191,13 +192,12 @@ unpack_Z_stream(transformer_aux_data_t *aux, int src_fd, int dst_fd)
 			}
 			{
 				unsigned char *p = &inbuf[posbits >> 3];
-
-				code = ((((long) (p[0])) | ((long) (p[1]) << 8) |
-				         ((long) (p[2]) << 16)) >> (posbits & 0x7)) & bitmask;
+				code = ((p[0]
+					| ((long) (p[1]) << 8)
+					| ((long) (p[2]) << 16)) >> (posbits & 0x7)) & bitmask;
 			}
 			posbits += n_bits;
 
-
 			if (oldcode == -1) {
 				if (code >= 256)
 					bb_error_msg_and_die("corrupted data"); /* %ld", code); */
@@ -226,15 +226,16 @@ unpack_Z_stream(transformer_aux_data_t *aux, int src_fd, int dst_fd)
 			/* Special case for KwKwK string. */
 			if (code >= free_ent) {
 				if (code > free_ent) {
+/*
 					unsigned char *p;
 
 					posbits -= n_bits;
 					p = &inbuf[posbits >> 3];
-
 					bb_error_msg
 						("insize:%d posbits:%d inbuf:%02X %02X %02X %02X %02X (%d)",
-						 insize, posbits, p[-1], p[0], p[1], p[2], p[3],
-						 (posbits & 07));
+						insize, posbits, p[-1], p[0], p[1], p[2], p[3],
+						(posbits & 07));
+*/
 					bb_error_msg("corrupted data");
 					goto err;
 				}
@@ -244,7 +245,7 @@ unpack_Z_stream(transformer_aux_data_t *aux, int src_fd, int dst_fd)
 			}
 
 			/* Generate output characters in reverse order */
-			while ((long) code >= (long) 256) {
+			while (code >= 256) {
 				if (stackp <= &htabof(0))
 					bb_error_msg_and_die("corrupted data");
 				*--stackp = tab_suffixof(code);
@@ -271,7 +272,7 @@ unpack_Z_stream(transformer_aux_data_t *aux, int src_fd, int dst_fd)
 						}
 
 						if (outpos >= OBUFSIZ) {
-							xwrite(dst_fd, outbuf, outpos);
+							xtransformer_write(xstate, outbuf, outpos);
 							IF_DESKTOP(total_written += outpos;)
 							outpos = 0;
 						}
@@ -285,21 +286,19 @@ unpack_Z_stream(transformer_aux_data_t *aux, int src_fd, int dst_fd)
 			}
 
 			/* Generate the new entry. */
-			code = free_ent;
-			if (code < maxmaxcode) {
-				tab_prefixof(code) = (unsigned short) oldcode;
-				tab_suffixof(code) = (unsigned char) finchar;
-				free_ent = code + 1;
+			if (free_ent < maxmaxcode) {
+				tab_prefixof(free_ent) = (unsigned short) oldcode;
+				tab_suffixof(free_ent) = (unsigned char) finchar;
+				free_ent++;
 			}
 
 			/* Remember previous code.  */
 			oldcode = incode;
 		}
-
 	} while (rsize > 0);
 
 	if (outpos > 0) {
-		xwrite(dst_fd, outbuf, outpos);
+		xtransformer_write(xstate, outbuf, outpos);
 		IF_DESKTOP(total_written += outpos;)
 	}
 
