@@ -6,6 +6,15 @@
  *
  * Licensed under GPLv2 or later, see file LICENSE in this source tree.
  */
+//config:config RMDIR
+//config:	bool "rmdir (3.5 kb)"
+//config:	default y
+//config:	help
+//config:	rmdir is used to remove empty directories.
+
+//applet:IF_RMDIR(APPLET_NOFORK(rmdir, rmdir, BB_DIR_BIN, BB_SUID_DROP, rmdir))
+
+//kbuild:lib-$(CONFIG_RMDIR) += rmdir.o
 
 /* BB_AUDIT SUSv3 compliant */
 /* http://www.opengroup.org/onlinepubs/007904975/utilities/rmdir.html */
@@ -14,12 +23,9 @@
 //usage:       "[OPTIONS] DIRECTORY..."
 //usage:#define rmdir_full_usage "\n\n"
 //usage:       "Remove DIRECTORY if it is empty\n"
-//usage:	IF_FEATURE_RMDIR_LONG_OPTIONS(
-//usage:     "\n	-p|--parents	Include parents"
-//usage:     "\n	--ignore-fail-on-non-empty"
-//usage:	)
-//usage:	IF_NOT_FEATURE_RMDIR_LONG_OPTIONS(
 //usage:     "\n	-p	Include parents"
+//usage:	IF_LONG_OPTS(
+//usage:     "\n	--ignore-fail-on-non-empty"
 //usage:	)
 //usage:
 //usage:#define rmdir_example_usage
@@ -30,8 +36,9 @@
 /* This is a NOFORK applet. Be very careful! */
 
 
-#define PARENTS 0x01
-#define IGNORE_NON_EMPTY 0x02
+#define PARENTS          (1 << 0)
+#define VERBOSE          ((1 << 1) * ENABLE_FEATURE_VERBOSE)
+#define IGNORE_NON_EMPTY ((1 << 2) * ENABLE_LONG_OPTS)
 
 int rmdir_main(int argc, char **argv) MAIN_EXTERNALLY_VISIBLE;
 int rmdir_main(int argc UNUSED_PARAM, char **argv)
@@ -40,16 +47,15 @@ int rmdir_main(int argc UNUSED_PARAM, char **argv)
 	int flags;
 	char *path;
 
-#if ENABLE_FEATURE_RMDIR_LONG_OPTIONS
-	static const char rmdir_longopts[] ALIGN1 =
+	flags = getopt32long(argv, "pv",
 		"parents\0"                  No_argument "p"
 		/* Debian etch: many packages fail to be purged or installed
 		 * because they desperately want this option: */
 		"ignore-fail-on-non-empty\0" No_argument "\xff"
-		;
-	applet_long_options = rmdir_longopts;
-#endif
-	flags = getopt32(argv, "p");
+		IF_FEATURE_VERBOSE(
+		"verbose\0"                  No_argument "v"
+		)
+	);
 	argv += optind;
 
 	if (!*argv) {
@@ -60,8 +66,12 @@ int rmdir_main(int argc UNUSED_PARAM, char **argv)
 		path = *argv;
 
 		while (1) {
+			if (flags & VERBOSE) {
+				printf("rmdir: removing directory, '%s'\n", path);
+			}
+
 			if (rmdir(path) < 0) {
-#if ENABLE_FEATURE_RMDIR_LONG_OPTIONS
+#if ENABLE_LONG_OPTS
 				if ((flags & IGNORE_NON_EMPTY) && errno == ENOTEMPTY)
 					break;
 #endif
